unit hiKnob;

interface

uses Windows,Kol,Share,Debug,Win,Messages;

type
  THIKnob = class(THIWin)
   private
    fcaption: string;
    fmarkmove: Boolean;  //  mousemove
    fposition: Integer;
    fMin: Integer;
    fMax: Integer;
    fPageSize: Word;
    fDiameter: Integer;
    fRadius: Integer;
    fSteps: Integer;
    fAngleInterval: Single;
    fAngle: Integer;
    fMouseAngle: Integer;
    fDragging: Boolean;
    fIndicatorInnerPosition: TPoint;
    fIndicatorOuterPosition: TPoint;
    fLastIndicatorInnerPosition: TPoint;
    fLastIndicatorOuterPosition: TPoint;

    procedure SetMin(const NewMinValue: Integer);
    procedure SetMax(const NewMaxValue: Integer);
    procedure SetParams(APosition, AMin, AMax, ASet: Integer);
    procedure SetSteps;
    procedure CalcAngle;
    function  CalcPosition(const TheAngle: Integer): Integer;
    property Position: Integer read fposition write fposition;

   protected
    function  _onMessage( var Msg: TMsg; var Rslt: Integer ): Boolean;
    procedure Paint( Sender: PControl; DC: HDC );
    procedure PaintKnob;
    procedure PaintState;
    procedure PaintIndicator;

   public
    _prop_MinValue:integer;
    _prop_MaxValue:integer;
    _prop_Position:integer;
    _prop_Steps:integer;
    _prop_PageSize:integer;
    _prop_CaptionVal:string;

    _event_onPosition:THI_Event;
    _event_onCaptionPos:THI_Event;

    destructor Destroy; override;
    procedure Init;override;
    procedure _work_doPosition(var _Data:TData; Index:word);
    procedure _work_doPosition2(var _Data:TData; Index:word);
    procedure _work_doCaptionVal(var _Data:TData; Index:word);
    procedure _work_MinValue(var _Data:TData; Index:word);
    procedure _work_MaxValue(var _Data:TData; Index:word);
    procedure _var_Position(var _Data:TData; Index:word);

    property MaxValue: Integer read fMax write SetMax;
    property MinValue: Integer read fMin write SetMin;
    property PageSize: Word read fPageSize write fPageSize default 10;
  end;

  function DegToRad(Degrees: Extended): Extended;
  function GetXYRadPoint2(AngleDegrees, Radius: Double; OffsetX, OffsetY : Double) : TPoint;
  procedure DrawGradientCircleRaised(Canvas: TCanvas; AColor: TColor; const CenterPoint: TPoint; Radius: Integer);
  procedure DrawGradientCircleSunken(Canvas: TCanvas; AColor: TColor; const CenterPoint: TPoint; Radius: Integer);
  procedure DrawGradientCircle(Canvas: TCanvas; StartColor, StopColor: TColor; const CenterPoint: TPoint; Radius: Integer; ReverseColors: Boolean);
  function ColorGradientBright  (AColor: TColor):TColor;
  function ColorGradientDim(AColor: TColor):TColor;
  function ColorOffset(AColor: TColor; AOffset: Byte; AIncrease: Boolean): TColor;

implementation

function DegToRad(Degrees: Extended): Extended;
begin
  Result := Degrees * (PI / 180);
end;

function ColorGradientBright(AColor: TColor):TColor;
begin
  Result := ColorOffset(AColor, $20, True );
end;

function ColorGradientDim(AColor: TColor):TColor;
begin
  Result := ColorOffset(AColor, $40, False);
end;

function ColorOffset(AColor: TColor; AOffset: Byte; AIncrease: Boolean): TColor;
var
  ARed, AGreen, ABlue : Integer;
begin
 AColor := Color2RGB(AColor);

 ARed   := (AColor and     $FF);
 AGreen := (AColor and   $FF00) shr 8;
 ABlue  := (AColor and $FF0000) shr 16;

 if AIncrease then
   begin
     ARed   := ARed   + AOffset;
     AGreen := AGreen + AOffset;
     ABlue  := ABlue  + AOffset;
   end
 else
   begin
     ARed   := ARed   - AOffset;
     AGreen := AGreen - AOffset;
     ABlue  := ABlue  - AOffset;
   end;

 if ARed   > 255 then ARed   := 255;
 if AGreen > 255 then AGreen := 255;
 if ABlue  > 255 then ABlue  := 255;

 if ARed   <   0 then ARed   := 0;
 if AGreen <   0 then AGreen := 0;
 if ABlue  <   0 then ABlue  := 0;

 Result := ARed + (AGreen shl 8) + (ABlue shl 16);
end;

function GetXYRadPoint2(AngleDegrees, Radius: Double; OffsetX, OffsetY : Double) : TPoint;
begin
  Result := MakePoint(Trunc(OffsetX + Cos(DegToRad(AngleDegrees))*Radius),Trunc(OffsetY - Sin(DegToRad(AngleDegrees))*Radius));
end;

procedure DrawGradientCircleRaised(Canvas: TCanvas; AColor: TColor; const CenterPoint: TPoint; Radius: Integer);
begin
  DrawGradientCircle(Canvas, ColorGradientBright(AColor), ColorGradientDim(AColor), CenterPoint, Radius, False);
end;

procedure DrawGradientCircleSunken(Canvas: TCanvas; AColor: TColor; const CenterPoint: TPoint; Radius: Integer);
begin
  DrawGradientCircle(Canvas, ColorGradientBright(AColor), ColorGradientDim(AColor), CenterPoint, Radius, True);
end;

procedure DrawGradientCircle(Canvas: TCanvas; StartColor, StopColor: TColor; const CenterPoint: TPoint; Radius: Integer; ReverseColors: Boolean);
var
  Red, Green, Blue : Integer;
  StartRed         : Integer;
  StartGreen       : Integer;
  StartBlue        : Integer;
  StopRed          : Integer;
  StopGreen        : Integer;
  StopBlue         : Integer;
  NumOfLines       : Integer;
  PointRadius      : Double;
  Size             : Double;
  Point1           : TPoint;
  Point2           : TPoint;
  LinePoint1       : TPoint;
  LinePoint2       : TPoint;
  MidX, MidY       : Double;
  PercentComplete  : Double;
begin
  with Canvas do
    begin
      StartColor := Color2RGB(StartColor);
      StopColor  := Color2RGB(StopColor);

      StartRed   := Byte(StartColor);
      StartGreen := Byte(StartColor shr 8);
      StartBlue  := Byte(StartColor shr 16);

      StopRed    := Byte(StopColor);
      StopGreen  := Byte(StopColor shr 8);
      StopBlue   := Byte(StopColor shr 16);

      Brush.BrushStyle := bsClear;
      Pen.PenStyle   := psSolid;

      NumOfLines := Radius*2;

      if NumOfLines = 0 then Exit;

      MidX := -Trunc(Radius/Sqrt(2)+2);
      MidY :=  Trunc(Radius/Sqrt(2)+2);

      while MidX < Radius/Sqrt(2)+2 do
        begin
          PointRadius := Sqrt(MidX*MidX + MidY*MidY);

          MidX := MidX + 0.5;
          MidY := MidY - 0.5;

          if PointRadius > Radius then Continue;

          Size := Sqrt(Radius*Radius - PointRadius*PointRadius);

          Point1 := GetXYRadPoint2(45 - 90, Size, MidX, MidY);
          Point2 := MakePoint(-Point1.Y, -Point1.X);

          PercentComplete := (MidX + Radius)/Radius/2;

          if ReverseColors then
            begin
              Red   := Round((PercentComplete*StartRed   + (1-PercentComplete)*StopRed  ));
              Green := Round((PercentComplete*StartGreen + (1-PercentComplete)*StopGreen));
              Blue  := Round((PercentComplete*StartBlue  + (1-PercentComplete)*StopBlue ));
            end
          else
            begin
              Red   := Round((PercentComplete*StopRed    + (1-PercentComplete)*StartRed  ));
              Green := Round((PercentComplete*StopGreen  + (1-PercentComplete)*StartGreen));
              Blue  := Round((PercentComplete*StopBlue   + (1-PercentComplete)*StartBlue ));
            end;

          Pen.Color := $02000000 or (Red + (Green shl 8) + (Blue shl 16));
          LinePoint1 := MakePoint(CenterPoint.X + Point1.X - 1, CenterPoint.Y - Point1.Y - 1);
          LinePoint2 := MakePoint(CenterPoint.X + Point2.X - 1, CenterPoint.Y - Point2.Y - 1);
          Polyline([LinePoint1, LinePoint2]);
        end;
    end;
end;


function THIKnob._onMessage;
var
  NewPosition: Integer;
  H:Integer; dt:Tdata;
begin
  Result:=false;
  case msg.message of
  WM_CREATE:
     begin
       SetSteps;
       CalcAngle;
     end;
  WM_SIZE:
     begin
       if (Control.Height > fDiameter) or (Control.Width > fDiameter)
         then H := Max(Control.Height, Control.Width) else
       if (Control.Height < fDiameter) or (Control.Width < fDiameter)
         then H := Min(Control.Height, Control.Width)
       else H := fDiameter;
       fDiameter := H;
       fRadius := H div 2;
    end;
  WM_SETFOCUS:
     begin
      Control.Focused := True;
      fDragging := True;
      //PaintState;
     end;
  WM_KILLFOCUS:
     begin
       Rslt := 1;
       //Control.Focused := False;
       //fDragging := False;
       //Control.Invalidate;
       //Paint(Control,Control.Canvas.Handle);
       Result:=True;
       PostMessage(Control.Handle,WM_PAINT,0,0);
       Exit;

     end;
  WM_MOUSEMOVE: begin
        fmarkmove := not fmarkmove;
        _hi_CreateEvent(dt,@_event_onPosition,fposition);
        if (msg.wParam = VK_LBUTTON) and (fmarkmove = true) then
        begin
          if loword(msg.lparam) = Control.Width div 2 then
          begin
            if hiword(msg.lParam) > Control.Height div 2
            then fMouseAngle := 240
            else fMouseAngle := 90;
          end
          else
          begin
            fMouseAngle := Round(ArcTan(((Control.Height div 2)-hiword(msg.lparam))/(loword(msg.lparam)-(Control.Width div 2)))*180/pi);
            if loword(msg.lparam) < Control.Width div 2 then fMouseAngle := (fMouseAngle + 540) MOD 360;
          end;

        if (fMouseAngle <= 270) and (fMouseAngle > 240) then fMouseAngle := 240;
        if (fMouseAngle < -60) then fMouseAngle := -60;

        NewPosition := CalcPosition(fMouseAngle);
        if fDragging and (fposition <> NewPosition) then Position:= NewPosition;
        SetParams(NewPosition,fMin,fMax,1);
        Control.Invalidate;
      end;
    end;
  WM_KEYDOWN:
     begin
       case msg.wparam  of
       //VK_PRIOR : begin Position:= Position + fPageSize; SetParams(Position,fMin,fMax,1);end;
       //VK_NEXT  : begin Position:= Position - fPageSize; SetParams(Position,fMin,fMax,1);end;
       VK_END   : begin Position:= fMin; SetParams(Position,fMin,fMax,1);end;
       VK_HOME  : begin Position:= fMax; SetParams(Position,fMin,fMax,1);end;
       VK_LEFT  : begin Position:= Position -1; SetParams(Position,fMin,fMax,1); Result:=True; end;
       VK_UP    : begin Position:= Position +1; SetParams(Position,fMin,fMax,1); Result:=True; end;
       VK_RIGHT : begin Position:= Position +1; SetParams(Position,fMin,fMax,1); Result:=True; end;
       VK_DOWN  : begin Position:= Position -1; SetParams(Position,fMin,fMax,1); Result:=True; end;
       end;
        //Control.Invalidate;
     end;
  WM_LBUTTONDOWN:
      begin
        Control.Focused := True;
        //SetFocus(Control.Handle);
        fDragging := True;
        PaintIndicator;
        Control.Invalidate;
        //_hi_onEvent(_event_onPosition,Position);
      end;
  WM_LBUTTONUP:
    if fDragging then
    begin
      _hi_onEvent(_event_onPosition,Position);
      fDragging := False;
    end;
  WM_MOUSEWHEEL:
    begin
      case msg.wparam  of   //const ???
         7864320: begin Position:= Position +1; SetParams(Position,fMin,fMax,1); Control.Invalidate;end;
        -7864320: begin Position:= Position -1; SetParams(Position,fMin,fMax,1); Control.Invalidate;end;
      end;
    end;
  else
    begin
      Result:=false;
    end;
  end;
end;

procedure THIKnob.Paint;
begin
  PaintKnob;
  PaintState;
  PaintIndicator;
end;

procedure THIKnob.SetMin(const NewMinValue: Integer);
begin
  SetParams(fposition, NewMinValue, fMax,1);
end;

procedure THIKnob.SetMax(const NewMaxValue: Integer);
begin
  SetParams(fposition, fMin, NewMaxValue,1);
end;

procedure THIKnob.SetParams(APosition, AMin, AMax, ASet: Integer);
begin
  if (fMin <> AMin) then
  begin
    fMin := AMin;
    SetSteps;
  end;

  if (fMax <> AMax) then
  begin
    fMax := AMax;
    SetSteps;
  end;

  if fAngleInterval >= 0 then
  begin
    if APosition < AMin then APosition := AMin;
    if APosition > AMax then APosition := AMax;
  end else
  begin
    if APosition > AMin then APosition := AMax;
    if APosition < AMax then APosition := AMin;
  end;

  if fposition <> APosition then Position := APosition;

  CalcAngle;
  PaintIndicator;
  if ASet = 1 then  _hi_onEvent(_event_onPosition,APosition);
  _hi_onEvent(_event_onCaptionPos, fcaption + int2str(APosition));
end;

procedure THIKnob.SetSteps;
begin
  fSteps := fMax - fMin;
  if fSteps = 0 then fAngleInterval:= 0 else
  begin
    fAngleInterval := 300 / fSteps;
    fSteps := Abs(fSteps);
  end;
end;

procedure THIKnob.CalcAngle;
begin
  fAngle := 240 - Round((fposition - fMin) * fAngleInterval);
end;

function THIKnob.CalcPosition(const TheAngle: Integer): Integer;
begin
  if fAngleInterval = 0 then Result := fMin else
    Result := fMin + Round((240 - TheAngle) / fAngleInterval);
end;

procedure THIKnob.PaintIndicator;
var
  AngleInRadians, CosAngle, SinAngle: Single;
begin
  AngleInRadians := fAngle*pi/180;
  CosAngle := Cos(AngleInRadians);
  SinAngle := Sin(AngleInRadians);

  with Control.Canvas{$ifndef F_P}^{$endif} do
  begin
    Pen.penwidth:=2;
    Pen.Color := Control.Color;
    with fIndicatorInnerPosition do MoveTo(X+1,Y+1);
    with fIndicatorOuterPosition do LineTo(X+1,Y+1);
    PolyLine([fIndicatorInnerPosition, fIndicatorOuterPosition]);

    fLastIndicatorInnerPosition := fIndicatorInnerPosition;
    fLastIndicatorOuterPosition := fIndicatorOuterPosition;

    fIndicatorInnerPosition.X := fRadius + Round(((fRadius)-11)*CosAngle);
    fIndicatorInnerPosition.Y := fRadius - Round(((fRadius)-11)*SinAngle);

    fIndicatorOuterPosition.X := fRadius + Round(((fRadius)-5)*CosAngle);
    fIndicatorOuterPosition.Y := fRadius - Round(((fRadius)-5)*SinAngle);

    if Control.Focused then
      Pen.Color := Control.Color2
    else
      Pen.Color := Control.Color1; //Control.Color2;
    with fIndicatorInnerPosition do MoveTo(X+1,Y+1);
    with fIndicatorOuterPosition do LineTo(X+1,Y+1);

    if Control.Focused then
      Pen.Color := Control.Color1
    else
      Pen.Color := Control.Color2;//Control.Color1;
    PolyLine([fIndicatorInnerPosition, fIndicatorOuterPosition]);
  end;
end;

procedure THIKnob.PaintState;
var
  R: TRect;
begin
  R := Control.ClientRect;
  with Control.Canvas{$ifndef F_P}^{$endif} do
  if Control.Focused then
  begin
    Pen.PenWidth:= 1;
    Brush.BrushStyle:= bsClear;
    if Control.Enabled then Pen.Color := Control.Color1 else Pen.Color := Control.Color2;//Control.Color1;
    if Control.focused then Pen.Color := Control.Color1 else Pen.Color := Control.Color2;//Control.Color1;
    Ellipse(0, 0, Control.Width, Control.Height);
    //Arc(R.Left,R.Top,R.Right,R.Bottom,R.Right,R.Top,R.Left,R.Bottom);
  end;
end;


procedure THIKnob.PaintKnob;
var
  R: TRect;
  CenterPoint, tmp: TPoint;
  Radius: integer;
begin
  R := Control.ClientRect;
  Radius := Control.Width div 2;
  CenterPoint.X := R.Left + Control.Width div 2;
  CenterPoint.Y := R.Top + Control.Height div 2;
  DrawGradientCircle(Control.Canvas{$ifndef F_P}^{$endif},Control.Color2,Control.Color1,CenterPoint,Radius,False);
end;

procedure THIKnob.Init;
begin
  Control := _NewControl(FParent,'THIKnob', WS_VISIBLE or WS_CHILD or BS_PUSHLIKE or WS_TABSTOP,False,nil);
  //inherited;
  fRadius := Control.width div 2;
  fDragging:=false;
  //fPageSize:= Pagesize;
  fPageSize:= 10;
  fMouseAngle:=240;
  fcaption:= _prop_CaptionVal;
  fMin:= _prop_MinValue;
  fMax:= _prop_MaxValue;
  Position:= _prop_Position;
  SetSteps;
  CalcAngle;
  SetParams(Position,fMin,fMax,1);
  Control.Transparent:=true;
  Control.OnPaint := Paint;
  Control.OnMessage := _OnMessage;
  inherited;
  Control.Color1:=ColorGradientDim(Control.Color);
  Control.Color2:=ColorGradientBright(Control.Color);
  Control.Invalidate;
end;

destructor THIKnob.Destroy;
begin
   inherited Destroy;
end;

procedure THIKnob._work_doPosition;
begin
  SetSteps;
  CalcAngle;
  SetParams(ToInteger(_Data),fMin,fMax,1);
end;

procedure THIKnob._work_doPosition2;
begin
  SetSteps;
  CalcAngle;
  SetParams(ToInteger(_Data),fMin,fMax,0);
end;

procedure THIKnob._work_MinValue;
begin
  fMin:= ToInteger(_Data);
  SetSteps;
  CalcAngle;
  SetParams(Position,fMin,fMax,1);
end;

procedure THIKnob._work_MaxValue;
begin
  fMax:= ToInteger(_Data);
  SetSteps;
  CalcAngle;
  SetParams(Position,fMin,fMax,1);
end;

procedure THIKnob._work_doCaptionVal;
begin
  fcaption:= ToString(_Data);
end;

procedure THIKnob._var_Position;
var dt: TData;
begin
  dtInteger(dt, Position);
end;

end.
